﻿using HoloToolkit.Unity;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

public class RealWorldObjectManager : Singleton<RealWorldObjectManager> {

    List<RealWorldObject> worldObjects;

    int realWorldObjectId = 0;

    bool isObjectsLoaded = false;

    [Tooltip("Anchor base name in disk")]
    public string realWorldObjectName = "cube";

    [Tooltip("the object to recognize location")]
    public GameObject markerObject;

    [Tooltip("distance of marker from real position")]
    public Vector3 distance;

    [Tooltip("Menu dismiss sound")]
    public AudioClip dismissMenuSound;

    // Use this for initialization
    void Start () {
        worldObjects = new List<RealWorldObject>();
    }

    protected override void Awake()
    {
        base.Awake();
    }

    public void dismissMenu()
    {
        if (dismissMenuSound != null)
        {
            AudioSource src = GetComponent<AudioSource>();
            src.clip = dismissMenuSound;
            src.Play();
        }
    }

    void Update () {

        
    }

    public void onRealWorldObjectsFound(List<RayPoints> rays)
    {
        float minDistanceDiffClasses = 0.3f;
        float minDistanceSameClasses = 0.3f;
        while (rays.Count != 0)
        {
            RayPoints rootRay = rays[0];
            HashSet<RayPoints> closedPoints = new HashSet<RayPoints>();
            for (int i = 0; i < rays.Count; i++)
            {
                float tmp = Vector3.Distance(rootRay.detectedObject.object3DPos,
                    rays[i].detectedObject.object3DPos);
                if (tmp <= minDistanceDiffClasses
                    || (rootRay.detectedObject.label == rays[i].detectedObject.label
                        && tmp <= minDistanceSameClasses))
                {
                    closedPoints.Add(rays[i]);
                }
            }
            int confidence = 0;
            RayPoints chosenPoint = null;
            foreach (RayPoints rp in closedPoints)
            {
                if (rp.detectedObject.score > confidence)
                {
                    confidence = rp.detectedObject.score;
                    chosenPoint = rp;
                }
            }
            rays.RemoveAll(s => closedPoints.Contains(s));
            addObject(chosenPoint.detectedObject.object3DPos,
                chosenPoint.detectedObject.label, chosenPoint.detectedObject.score);
        }
        writeObjectsToDisk();
    }

    public GameObject getClosesObject(Vector3 pos, float threshold)
    {
        float minDis = 100.0f;
        GameObject suspect = null;
        foreach (var worldObject in worldObjects)
        {
            float distFromBox = Vector3.Distance(worldObject.gameObj.transform.position, pos);
            if (distFromBox <= minDis && distFromBox <= threshold)
            {
                suspect = worldObject.gameObj;
                minDis = distFromBox;
            }
        }
        return suspect;
    }

    public GameObject addObject(Vector3 pos, string label, int confidence = 100)
    {
        var cube = getMarkerObject();
        cube.GetComponent<Interactible>().speakOnTap = label;
        cube.transform.position = pos + this.distance;
        worldObjects.Add(new RealWorldObject()
        {
            name = realWorldObjectName + realWorldObjectId.ToString(),
            gameObj = cube,
            label = label,
            confidence = confidence
        });
        WorldAnchorManager.Instance.AttachAnchor(cube, realWorldObjectName + realWorldObjectId.ToString());
        realWorldObjectId++;
        return cube;
    }

    public IEnumerator removeObject(GameObject anchor)
    {
        RealWorldObject obj = null;
        foreach (var worldObj in worldObjects)
        {
            if (worldObj.gameObj == anchor)
            {
                obj = worldObj;
                break;
            }
        }
        if (obj == null)
        {
            MySocket.log("didnt find object");
            yield return null;
        }
        WorldAnchorManager.Instance.RemoveAnchor(anchor);
        yield return new WaitUntil(() => WorldAnchorManager.Instance.AnchorOperationCount() == 0);
        worldObjects.Remove(obj);
        writeObjectsToDisk();
        Destroy(anchor);
    }

    public IEnumerator clean(System.Action act)
    {
        foreach (var worldObj in worldObjects)
        {
            WorldAnchorManager.Instance.RemoveAnchor(worldObj.gameObj);
            yield return new WaitUntil(() => WorldAnchorManager.Instance.AnchorOperationCount() == 0);
            Destroy(worldObj.gameObj);
        }
        worldObjects.Clear();
        writeObjectsToDisk();
        act();
    }

    public IEnumerator changeObjectLocation(GameObject anchor, Vector3 newPos)
    {
        RealWorldObject obj = null;
        foreach (var worldObj in worldObjects)
        {
            if (worldObj.gameObj == anchor) {
                obj = worldObj;
                break;
            }
        }
        if (obj == null)
        {
            MySocket.log("didnt find object");
            yield return null;
        }
        WorldAnchorManager.Instance.RemoveAnchor(anchor);
        MySocket.log("Removing");
        yield return new WaitUntil(() => WorldAnchorManager.Instance.AnchorOperationCount() == 0);
        MySocket.log("Removing done !");
        anchor.transform.position = newPos;
        WorldAnchorManager.Instance.AttachAnchor(anchor, obj.name);
    }

    public GameObject addObject(Vector3 pos, string label, Vector3 diff)
    {
        return this.addObject(pos + diff, label);
    }

        public void writeObjectsToDisk()
    {
        Dictionary<string, List<string>> dict = new Dictionary<string, List<string>>();
        foreach (var realWorldObj in worldObjects)
        {
            dict.Add(realWorldObj.name, new List<string>() { realWorldObj.label, realWorldObj.confidence.ToString() });
        }
        FileObjectsManager.Instance.writeToFileManager(dict);
    }

    public void onObjectsLoadedFromDisk(Dictionary<string, List<string>> loadedObjects)
    {
        foreach (string objectName in loadedObjects.Keys)
        {
            var cube = getMarkerObject();
            cube.GetComponent<Interactible>().speakOnTap = loadedObjects[objectName][0];
            worldObjects.Add(new RealWorldObject()
            {
                gameObj = cube,
                name = objectName,
                confidence = int.Parse(loadedObjects[objectName][1]),
                label = loadedObjects[objectName][0]
            });
            WorldAnchorManager.Instance.AttachAnchor(cube, objectName);
        }
        if (loadedObjects.Keys.Count > 0)
        {
            this.realWorldObjectId = loadedObjects.Keys.Max(s => int.Parse(s.Replace("cube", "")));
            this.realWorldObjectId++;
        }
        isObjectsLoaded = true;
    }

    public List<string> getWorldObjects() {
        return worldObjects.Select(s => s.label).ToList();
    }

    public string getWorldObject(GameObject gameObj)
    {
        return worldObjects.Where(s => s.gameObj == gameObj).Select(s => s.label)
            .FirstOrDefault();
    }

    public bool isWorldObjectsEmpty()
    {
        return worldObjects.Count == 0;
    }

    public List<GameObject> getGameObjectsWhere(string label)
    {
        return worldObjects.Where(s => s.label == label).Select(s => s.gameObj).ToList();
    }

    public void showWorldObjects()
    {
        foreach (var worldObj in this.worldObjects)
        {
            showWorldObject(worldObj.gameObj, false);
        }
    }

    public void hideWorldObjects()
    {
        foreach (var worldObj in this.worldObjects)
        {
            hideWorldObject(worldObj.gameObj);
        }
    }

    public void hideWorldObject(GameObject gameObj)
    {
        if (gameObj.GetComponentInChildren<DirectionIndicator>() != null)
        {
            gameObj.GetComponentInChildren<DirectionIndicator>().enabled = false;
        }
        foreach (var cmp in gameObj.GetComponentsInChildren<MeshRenderer>(true))
        {
            cmp.enabled = false;
        }
    }

    public void showWorldObject(GameObject gameObj, bool isNavigation)
    {
        if (!gameObj.activeSelf)
        {
            gameObj.SetActive(true);
        }
        if (gameObj.GetComponentInChildren<DirectionIndicator>() != null)
        {
            gameObj.GetComponentInChildren<DirectionIndicator>().enabled = isNavigation;
        }
        foreach (var cmp in gameObj.GetComponentsInChildren<MeshRenderer>(true))
        {
            cmp.enabled = true;
        }
    }

    GameObject getMarkerObject()
    {
        if (markerObject != null)
        {
            var marker = Instantiate(this.markerObject);
            marker.SetActive(true);
            hideWorldObject(marker);
            return marker;
        }
        else
        {
            var cube = GameObject.CreatePrimitive(PrimitiveType.Cylinder);
            cube.transform.localScale = Vector3.one * 0.1f;
            cube.GetComponent<Renderer>().enabled = false;
            cube.AddComponent<Interactible>();
            return cube;
        }
    }

    private void OnApplicationQuit()
    {
        MySocket.closeConnection();
    }
}
public class RealWorldObject
{
    public GameObject gameObj;
    public string name;
    public string label;
    public int confidence;
}
