import socket
import os
import json
import base64
from enum import Enum
from manager import ImageManager
from object2d import ImageFrame
from run import CMTTracker
import datetime
import cv2

class State(Enum):
    IMAGE = 1
    TRACKING = 2

class MySocket:

    TMP_DIR = 'tmp'
    TMP_TRACKER = 'tmp_tracker'
    HOST = '132.68.34.36'
    PORT = 20605
    IMAGE_BUFFER_SIZE = 4096
    TEXT_BUFFER_SIZE  = 256
    MAX_CONNNECTIONS  = 1
    IMAGE_FORMAT = '.jpeg'

    IMAGE_KEY = "image_file"
    ID_KEY = "id"
    LAST_KEY = "last"
    OPCODE_KEY = "opcode"

    def __init__(self):
        self.state = State.IMAGE
        self.processing_mode = False
        self.imageMgr = ImageManager()

    def run(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind((MySocket.HOST, MySocket.PORT))
        s.listen(MySocket.MAX_CONNNECTIONS)
        print("Listening on {} with port {}".format(MySocket.HOST, MySocket.PORT))
        while True:
            client_socket, address = s.accept()
            self.handleClientSocket(client_socket)
        s.close()

    def RGBA32BytesIntoImage(self, im_bytes, file_name):
        f = open(file_name, 'wb')
        f.write(im_bytes)
        f.close()

    def validateFrameJson(self, data):
        if MySocket.IMAGE_KEY not in data.keys() or\
            MySocket.LAST_KEY not in data.keys() or \
            MySocket.ID_KEY not in data.keys():
            raise RuntimeError("Wrong json object.")

    def __recieveBytes(self, client_socket, msg_len):
        chunks = []
        bytes_recd = 0
        while bytes_recd < msg_len:
            chunk = client_socket.recv(min(msg_len - bytes_recd, 2048))
            if chunk == b'':
                raise RuntimeError("socket connection broken")
            chunks.append(chunk)
            bytes_recd = bytes_recd + len(chunk)
        return b''.join(chunks)

    def __sendBytes(self, client_socket, msg):
        totalsent = 0
        while totalsent < len(msg):
            sent = client_socket.send(msg[totalsent:])
            if sent == 0:
                raise RuntimeError("socket connection broken")
            totalsent = totalsent + sent

    def readMsg(self, client_socket):
        data = self.__recieveBytes(client_socket, msg_len = 4)
        msg_len = int.from_bytes(data, byteorder='little', signed=True)
        return self.__recieveBytes(client_socket, msg_len)

    def writeMsg(self, client_socket, msg):
        print(len(msg))
        self.__sendBytes(client_socket, len(msg).to_bytes(4, byteorder='little', signed=True))
        self.__sendBytes(client_socket, msg)

    def recieveImage(self, client_socket, data):
        if not os.path.exists(MySocket.TMP_DIR):
            os.makedirs(MySocket.TMP_DIR)
        data = data.decode('utf-8')
        data = json.loads(data, encoding = 'utf-8')
        self.validateFrameJson(data)
        if data[MySocket.LAST_KEY]:
            self.processing_mode = True
            return
        decodedImg = base64.b64decode(data[MySocket.IMAGE_KEY])
        img_path = os.path.join(MySocket.TMP_DIR, 'image' + data[MySocket.ID_KEY] + MySocket.IMAGE_FORMAT)
        self.RGBA32BytesIntoImage(decodedImg, img_path)
        frame = ImageFrame(img_path, data[MySocket.ID_KEY])
        self.imageMgr.add_frame(frame)
        print(datetime.datetime.now())
        print("Got Image Frame")

    def recieveImageForTracking(self, client_socket, data):
        if not os.path.exists(MySocket.TMP_TRACKER):
            os.makedirs(MySocket.TMP_TRACKER)
        data = data.decode('utf-8')
        data = json.loads(data, encoding = 'utf-8')
        decodedImg = base64.b64decode(data[MySocket.IMAGE_KEY])
        img_path = os.path.join(MySocket.TMP_TRACKER, 'image' + data[MySocket.ID_KEY] + MySocket.IMAGE_FORMAT)
        self.RGBA32BytesIntoImage(decodedImg, img_path)
        if 'x' in data and 'y' in data:
            print("Got first image " + str(img_path))
            print((int(data['x']), int(data['y'])))
            center = (int(data['x']), int(data['y']))
            try:
                self.tracker = CMTTracker((int(data['x']), int(data['y'])), cv2.imread(img_path))
            except Exception as e:
                print(str(e))
                center = (0, 0)
        else:
            try:
                center = self.tracker.get_next_box(cv2.imread(img_path))
            except Exception as e:
                print(str(e))
                center = (0, 0)
        center = (str(int(center[0])), str(int(center[1])))
        dict = {"x": center[0], "y": center[1]}
        str1 = json.dumps(dict)
        str2 = str1.encode('utf-8')
        self.writeMsg(client_socket, str2)
        print(datetime.datetime.now())
        print("Got Image Frame")

    def reset(self):
        self.state = State.IMAGE
        self.processing_mode = False
        self.imageMgr = ImageManager()

    def get_state(self, data):
        data = data.decode('utf-8')
        data = json.loads(data, encoding='utf-8')
        if data[MySocket.OPCODE_KEY] == "SCAN":
            return State.IMAGE
        if data[MySocket.OPCODE_KEY] == "TRACK":
            return State.TRACKING
        return State.IMAGE

    def handleClientSocket(self, client_socket):
        print("Got Client")
        while True:
            try:
                msg = self.readMsg(client_socket)
                self.state = self.get_state(msg)
                if self.state == State.IMAGE:
                    self.recieveImage(client_socket, msg)
                elif self.state == State.TRACKING:
                    self.recieveImageForTracking(client_socket, msg)
                if self.processing_mode:
                    print("Start processing images")
                    self.imageMgr.process_images()
                    dict = {}
                    for frame in self.imageMgr.frames:
                        dict[frame.image_id] = frame.boxes
                    print(dict)
                    str1 = json.dumps(dict)
                    str2 = str1.encode('utf-8')
                    self.writeMsg(client_socket, str2)
                    print("Done processing, you can ask questions now")
                    self.imageMgr.frames.clear()
                    self.processing_mode = False
                    print("Done")
            except OSError:
                client_socket.close()
                print("Connection closed")
                break
        self.reset()

if __name__ == '__main__':
    server = MySocket()
    server.run()

